@@ -1,58 +1,111 @@
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE TypeSynonymInstances #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE PackageImports #-}
 {-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 
--- | This defines the API for performing database actions. There are two levels
--- to this API: dealing with fields, and dealing with entities. In SQL, a field
--- corresponds to a column, and should be a single, non-composite value. An
--- entity corresponds to a SQL table.  In other words: An entity is a
--- collection of fields.
+-- | API for database actions. The API deals with fields and entities.
+-- In SQL, a field corresponds to a column, and should be a single non-composite value.
+-- An entity corresponds to a SQL table, so an entity is a collection of fields.
 module Database.Persist.Base
     ( PersistValue (..)
     , SqlType (..)
     , PersistField (..)
     , PersistEntity (..)
-    , EntityDef (..)
     , PersistBackend (..)
     , PersistFilter (..)
     , PersistUpdate (..)
-    , PersistOrder (..)
+    , SelectOpt (..)
     , SomePersistField (..)
+
     , selectList
     , insertBy
+    , getByValue
+    , getJust
+    , belongsTo
+    , belongsToJust
+
     , checkUnique
     , DeleteCascade (..)
     , deleteCascadeWhere
     , PersistException (..)
+    , Update (..)
+    , updateFieldName
+    , Filter (..)
+    , Key (..)
+      -- * Definition
+    , EntityDef (..)
+    , ColumnName
+    , ColumnType
+    , ColumnDef (..)
+    , UniqueDef (..)
+    , fst3, snd3, third3
+    , limitOffsetOrder
+
+      -- * Config
+    , PersistConfig (..)
     ) where
 
-import Language.Haskell.TH.Syntax
 import Data.Time (Day, TimeOfDay, UTCTime)
 import Data.ByteString.Char8 (ByteString, unpack)
 import Control.Applicative
 import Data.Typeable (Typeable)
 import Data.Int (Int8, Int16, Int32, Int64)
 import Data.Word (Word8, Word16, Word32, Word64)
-import Text.Hamlet
+import Text.Blaze (Html, unsafeByteString)
+import Text.Blaze.Renderer.Utf8 (renderHtml)
 import qualified Data.Text as T
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
-import Data.Enumerator
+import Data.Enumerator hiding (consume, map)
+import Data.Enumerator.List (consume)
+import qualified Data.Enumerator.List as EL
+import qualified Control.Monad.IO.Class as Trans
+
 import qualified Control.Exception as E
+import Control.Monad.Error.Class (Error (..))
+
 import Data.Bits (bitSize)
 import Control.Monad (liftM)
 
 import qualified Data.Text.Encoding as T
 import qualified Data.Text.Encoding.Error as T
+import Web.PathPieces (SinglePiece (..))
+import qualified Data.Text.Read
+import Data.Maybe (fromMaybe, isJust)
+
+import Control.Monad.IO.Control (MonadControlIO)
+import Data.Object (TextObject)
+
+fst3 :: forall t t1 t2. (t, t1, t2) -> t
+fst3   (x, _, _) = x
+snd3 :: forall t t1 t2. (t, t1, t2) -> t1
+snd3   (_, x, _) = x
+third3 :: forall t t1 t2. (t, t1, t2) -> t2
+third3 (_, _, x) = x
+
+data PersistException
+  = PersistError String -- ^ Generic Exception
+  | PersistMarshalError String
+  | PersistInvalidField String
+  | PersistForeignConstraintUnmet String
+  | PersistMongoDBError String
+  | PersistMongoDBUnsupported String
+    deriving (Show, Typeable)
+
+instance E.Exception PersistException
+instance Error PersistException where strMsg msg = PersistError msg
+
+
 
 -- | A raw value which can be stored in any backend and can be marshalled to
 -- and from a 'PersistField'.
-data PersistValue = PersistString String
+data PersistValue = PersistText T.Text
                   | PersistByteString ByteString
                   | PersistInt64 Int64
                   | PersistDouble Double
@@ -61,14 +114,28 @@ data PersistValue = PersistString String
                   | PersistTimeOfDay TimeOfDay
                   | PersistUTCTime UTCTime
                   | PersistNull
-    deriving (Show, Read, Eq, Typeable)
+                  | PersistList [PersistValue]
+                  | PersistMap [(T.Text, PersistValue)]
+                  | PersistObjectId ByteString -- ^ intended especially for MongoDB backend
+    deriving (Show, Read, Eq, Typeable, Ord)
+
+instance SinglePiece PersistValue where
+    fromSinglePiece t =
+        case Data.Text.Read.signed Data.Text.Read.decimal t of
+            Right (i, t')
+                | T.null t' -> Just $ PersistInt64 i
+            _ -> Just $ PersistText t
+    toSinglePiece x =
+        case fromPersistValue x of
+            Left e -> error e
+            Right y -> y
 
 -- | A SQL data type. Naming attempts to reflect the underlying Haskell
 -- datatypes, eg SqlString instead of SqlVarchar. Different SQL databases may
 -- have different translations for these types.
 data SqlType = SqlString
              | SqlInt32
-             | SqlInteger -- ^ 8-byte integer; should be renamed SqlInt64
+             | SqlInteger -- ^ FIXME 8-byte integer; should be renamed SqlInt64
              | SqlReal
              | SqlBool
              | SqlDay
@@ -86,8 +153,8 @@ class PersistField a where
     isNullable _ = False
 
 instance PersistField String where
-    toPersistValue = PersistString
-    fromPersistValue (PersistString s) = Right s
+    toPersistValue = PersistText . T.pack
+    fromPersistValue (PersistText s) = Right $ T.unpack s
     fromPersistValue (PersistByteString bs) =
         Right $ T.unpack $ T.decodeUtf8With T.lenientDecode bs
     fromPersistValue (PersistInt64 i) = Right $ show i
@@ -97,19 +164,32 @@ instance PersistField String where
     fromPersistValue (PersistUTCTime d) = Right $ show d
     fromPersistValue PersistNull = Left "Unexpected null"
     fromPersistValue (PersistBool b) = Right $ show b
+    fromPersistValue (PersistList _) = Left "Cannot convert PersistList to String"
+    fromPersistValue (PersistMap _) = Left "Cannot convert PersistMap to String"
+    fromPersistValue (PersistObjectId _) = Left "Cannot convert PersistObjectId to String"
     sqlType _ = SqlString
 
 instance PersistField ByteString where
     toPersistValue = PersistByteString
     fromPersistValue (PersistByteString bs) = Right bs
-    fromPersistValue x = T.encodeUtf8 . T.pack <$> fromPersistValue x
+    fromPersistValue x = T.encodeUtf8 <$> fromPersistValue x
     sqlType _ = SqlBlob
 
 instance PersistField T.Text where
-    toPersistValue = PersistString . T.unpack
+    toPersistValue = PersistText
+    fromPersistValue (PersistText s) = Right s
     fromPersistValue (PersistByteString bs) =
         Right $ T.decodeUtf8With T.lenientDecode bs
-    fromPersistValue v = fmap T.pack $ fromPersistValue v
+    fromPersistValue (PersistInt64 i) = Right $ T.pack $ show i
+    fromPersistValue (PersistDouble d) = Right $ T.pack $ show d
+    fromPersistValue (PersistDay d) = Right $ T.pack $ show d
+    fromPersistValue (PersistTimeOfDay d) = Right $ T.pack $ show d
+    fromPersistValue (PersistUTCTime d) = Right $ T.pack $ show d
+    fromPersistValue PersistNull = Left "Unexpected null"
+    fromPersistValue (PersistBool b) = Right $ T.pack $ show b
+    fromPersistValue (PersistList _) = Left "Cannot convert PersistList to Text"
+    fromPersistValue (PersistMap _) = Left "Cannot convert PersistMap to Text"
+    fromPersistValue (PersistObjectId _) = Left "Cannot convert PersistObjectId to Text"
     sqlType _ = SqlString
 
 instance PersistField Html where
@@ -189,8 +269,8 @@ instance PersistField Bool where
 instance PersistField Day where
     toPersistValue = PersistDay
     fromPersistValue (PersistDay d) = Right d
-    fromPersistValue x@(PersistString s) =
-        case reads s of
+    fromPersistValue x@(PersistText t) =
+        case reads $ T.unpack t of
             (d, _):_ -> Right d
             _ -> Left $ "Expected Day, received " ++ show x
     fromPersistValue x@(PersistByteString s) =
@@ -203,8 +283,8 @@ instance PersistField Day where
 instance PersistField TimeOfDay where
     toPersistValue = PersistTimeOfDay
     fromPersistValue (PersistTimeOfDay d) = Right d
-    fromPersistValue x@(PersistString s) =
-        case reads s of
+    fromPersistValue x@(PersistText t) =
+        case reads $ T.unpack t of
             (d, _):_ -> Right d
             _ -> Left $ "Expected TimeOfDay, received " ++ show x
     fromPersistValue x@(PersistByteString s) =
@@ -217,8 +297,8 @@ instance PersistField TimeOfDay where
 instance PersistField UTCTime where
     toPersistValue = PersistUTCTime
     fromPersistValue (PersistUTCTime d) = Right d
-    fromPersistValue x@(PersistString s) =
-        case reads s of
+    fromPersistValue x@(PersistText t) =
+        case reads $ T.unpack t of
             (d, _):_ -> Right d
             _ -> Left $ "Expected UTCTime, received " ++ show x
     fromPersistValue x@(PersistByteString s) =
@@ -236,46 +316,50 @@ instance PersistField a => PersistField (Maybe a) where
     sqlType _ = sqlType (error "this is the problem" :: a)
     isNullable _ = True
 
+data Update v = forall typ. PersistField typ => Update
+    { updateField :: EntityField v typ
+    , updateValue :: typ
+    , updateUpdate :: PersistUpdate -- FIXME Replace with expr down the road
+    }
+
+updateFieldName :: PersistEntity v => Update v -> String
+updateFieldName (Update f _ _) = columnName $ persistColumnDef f
+
+data SelectOpt v = forall typ. Asc (EntityField v typ)
+                 | forall typ. Desc (EntityField v typ)
+                 | OffsetBy Int
+                 | LimitTo Int
+
+-- | Filters which are available for 'select', 'updateWhere' and
+-- 'deleteWhere'. Each filter constructor specifies the field being
+-- filtered on, the type of comparison applied (equals, not equals, etc)
+-- and the argument for the comparison.
+data Filter v = forall typ. PersistField typ => Filter
+    { filterField :: EntityField v typ
+    , filterValue :: Either typ [typ] -- FIXME
+    , filterFilter :: PersistFilter -- FIXME
+    }
+    | FilterAnd [Filter v] -- ^ convenient for internal use, not needed for the API
+    | FilterOr [Filter v]
+
 -- | A single database entity. For example, if writing a blog application, a
 -- blog entry would be an entry, containing fields such as title and content.
 class PersistEntity val where
-    -- | The unique identifier associated with this entity. In general, backends also define a type synonym for this, such that \"type MyEntityId = Key MyEntity\".
-    data Key    val
-    -- | Fields which can be updated using the 'update' and 'updateWhere'
-    -- functions.
-    data Update val
-    -- | Filters which are available for 'select', 'updateWhere' and
-    -- 'deleteWhere'. Each filter constructor specifies the field being
-    -- filtered on, the type of comparison applied (equals, not equals, etc)
-    -- and the argument for the comparison.
-    data Filter val
-    -- | How you can sort the results of a 'select'.
-    data Order  val
+    -- | Parameters: val and datatype of the field
+    data EntityField val :: * -> *
+    persistColumnDef :: EntityField val typ -> ColumnDef
+
     -- | Unique keys in existence on this entity.
-    data Unique val
+    data Unique val :: ((* -> *) -> * -> *) -> *
 
     entityDef :: val -> EntityDef
     toPersistFields :: val -> [SomePersistField]
     fromPersistValues :: [PersistValue] -> Either String val
     halfDefined :: val
-    toPersistKey :: Int64 -> Key val
-    fromPersistKey :: Key val -> Int64
-    showPersistKey :: Key val -> String
 
-    persistFilterToFieldName :: Filter val -> String
-    persistFilterToFilter :: Filter val -> PersistFilter
-    persistFilterToValue :: Filter val -> Either PersistValue [PersistValue]
-
-    persistOrderToFieldName :: Order val -> String
-    persistOrderToOrder :: Order val -> PersistOrder
-
-    persistUpdateToFieldName :: Update val -> String
-    persistUpdateToUpdate :: Update val -> PersistUpdate
-    persistUpdateToValue :: Update val -> PersistValue
-
-    persistUniqueToFieldNames :: Unique val -> [String]
-    persistUniqueToValues :: Unique val -> [PersistValue]
-    persistUniqueKeys :: val -> [Unique val]
+    persistUniqueToFieldNames :: Unique val backend -> [String]
+    persistUniqueToValues :: Unique val backend -> [PersistValue]
+    persistUniqueKeys :: val -> [Unique val backend]
 
 data SomePersistField = forall a. PersistField a => SomePersistField a
 instance PersistField SomePersistField where
@@ -283,60 +367,71 @@ instance PersistField SomePersistField where
     fromPersistValue x = fmap SomePersistField (fromPersistValue x :: Either String String)
     sqlType (SomePersistField a) = sqlType a
 
-class Monad m => PersistBackend m where
+newtype Key backend entity = Key { unKey :: PersistValue }
+    deriving (Show, Read, Eq, Ord, PersistField)
+
+class (Trans.MonadIO (b m), Trans.MonadIO m, Monad (b m), Monad m) => PersistBackend b m where
+
     -- | Create a new record in the database, returning the newly created
     -- identifier.
-    insert :: PersistEntity val => val -> m (Key val)
+    insert :: PersistEntity val => val -> b m (Key b val)
 
     -- | Replace the record in the database with the given key. Result is
     -- undefined if such a record does not exist.
-    replace :: PersistEntity val => Key val -> val -> m ()
+    replace :: PersistEntity val => Key b val -> val -> b m ()
 
     -- | Update individual fields on a specific record.
-    update :: PersistEntity val => Key val -> [Update val] -> m ()
+    update :: PersistEntity val => Key b val -> [Update val] -> b m ()
 
     -- | Update individual fields on any record matching the given criterion.
-    updateWhere :: PersistEntity val => [Filter val] -> [Update val] -> m ()
+    updateWhere :: PersistEntity val => [Filter val] -> [Update val] -> b m ()
 
     -- | Delete a specific record by identifier. Does nothing if record does
     -- not exist.
-    delete :: PersistEntity val => Key val -> m ()
+    delete :: PersistEntity val => Key b val -> b m ()
 
     -- | Delete a specific record by unique key. Does nothing if no record
     -- matches.
-    deleteBy :: PersistEntity val => Unique val -> m ()
+    deleteBy :: PersistEntity val => Unique val b -> b m ()
 
     -- | Delete all records matching the given criterion.
-    deleteWhere :: PersistEntity val => [Filter val] -> m ()
+    deleteWhere :: PersistEntity val => [Filter val] -> b m ()
 
     -- | Get a record by identifier, if available.
-    get :: PersistEntity val => Key val -> m (Maybe val)
+    get :: PersistEntity val => Key b val -> b m (Maybe val)
 
     -- | Get a record by unique key, if available. Returns also the identifier.
-    getBy :: PersistEntity val => Unique val -> m (Maybe (Key val, val))
+    getBy :: PersistEntity val => Unique val b -> b m (Maybe (Key b val, val))
 
     -- | Get all records matching the given criterion in the specified order.
     -- Returns also the identifiers.
-    select :: PersistEntity val
+    selectEnum
+           :: PersistEntity val
            => [Filter val]
-           -> [Order val]
-           -> Int -- ^ limit
-           -> Int -- ^ offset
-           -> Enumerator (Key val, val) m a
+           -> [SelectOpt val]
+           -> Enumerator (Key b val, val) (b m) a
+
+    -- | get just the first record for the criterion
+    selectFirst :: PersistEntity val
+                => [Filter val]
+                -> [SelectOpt val]
+                -> b m (Maybe (Key b val, val))
+    selectFirst filts opts = run_ $ selectEnum filts ((LimitTo 1):opts) ==<< EL.head
+
 
     -- | Get the 'Key's of all records matching the given criterion.
     selectKeys :: PersistEntity val
                => [Filter val]
-               -> Enumerator (Key val) m a
+               -> Enumerator (Key b val) (b m) a
 
     -- | The total number of records fulfilling the given criterion.
-    count :: PersistEntity val => [Filter val] -> m Int
+    count :: PersistEntity val => [Filter val] -> b m Int
 
 -- | Insert a value, checking for conflicts with any unique constraints.  If a
 -- duplicate exists in the database, it is returned as 'Left'. Otherwise, the
 -- new 'Key' is returned as 'Right'.
-insertBy :: (PersistEntity v, PersistBackend m)
-          => v -> m (Either (Key v, v) (Key v))
+insertBy :: (PersistEntity v, PersistBackend b m)
+          => v -> b m (Either (Key b v, v) (Key b v))
 insertBy val =
     go $ persistUniqueKeys val
   where
@@ -347,12 +442,53 @@ insertBy val =
             Nothing -> go xs
             Just z -> return $ Left z
 
+-- | A modification of 'getBy', which takes the 'PersistEntity' itself instead
+-- of a 'Unique' value. Returns a value matching /one/ of the unique keys. This
+-- function makes the most sense on entities with a single 'Unique'
+-- constructor.
+getByValue :: (PersistEntity v, PersistBackend b m)
+           => v -> b m (Maybe (Key b v, v))
+getByValue val =
+    go $ persistUniqueKeys val
+  where
+    go [] = return Nothing
+    go (x:xs) = do
+        y <- getBy x
+        case y of
+            Nothing -> go xs
+            Just z -> return $ Just z
+
+-- | curry this to make a convenience function that loads an associated model
+--   > foreign = belongsTo foeignId
+belongsTo ::
+  (PersistBackend b m
+  , PersistEntity ent1
+  , PersistEntity ent2) => (ent1 -> Maybe (Key b ent2)) -> ent1 -> b m (Maybe ent2)
+belongsTo foreignKeyField model = case foreignKeyField model of
+    Nothing -> return Nothing
+    Just f -> get f
+
+-- | same as belongsTo, but uses @getJust@ and therefore is similarly unsafe
+belongsToJust ::
+  (PersistBackend b m
+  , PersistEntity ent1
+  , PersistEntity ent2) => (ent1 -> Key b ent2) -> ent1 -> b m ent2
+belongsToJust getForeignKey model = getJust $ getForeignKey model
+
+-- | Same as get, but for a non-null (not Maybe) foreign key
+--   Unsafe unless your database is enforcing that the foreign key is valid
+getJust :: (PersistBackend b m, PersistEntity val, Show (Key b val)) => Key b val -> b m val
+getJust key = get key >>= maybe
+  (Trans.liftIO . E.throwIO $ PersistForeignConstraintUnmet (show key))
+  return
+
+
 -- | Check whether there are any conflicts for unique keys with this entity and
 -- existing entities in the database.
 --
 -- Returns 'True' if the entity would be unique, and could thus safely be
 -- 'insert'ed; returns 'False' on a conflict.
-checkUnique :: (PersistEntity val, PersistBackend m) => val -> m Bool
+checkUnique :: (PersistEntity val, PersistBackend b m) => val -> b m Bool
 checkUnique val =
     go $ persistUniqueKeys val
   where
@@ -363,67 +499,66 @@ checkUnique val =
             Nothing -> go xs
             Just _ -> return False
 
+limitOffsetOrder :: PersistEntity val => [SelectOpt val] -> (Int, Int, [SelectOpt val])
+limitOffsetOrder opts = let (l,o,ord) = go opts (Nothing, Nothing, []) in
+    (fromMaybe 0 l, fromMaybe 0 o, ord)
+  where
+    go []              tup = tup
+    go (LimitTo  x:xs) (_,o,ord) = let tup = (Just x, o, ord) in
+      if isJust o then tup else go xs tup
+    go (OffsetBy x:xs) (l,_,ord) = let tup = (l, Just x, ord) in
+      if isJust l then tup else go xs tup
+    go (x:xs)          (l, o, ord) = go xs (l,o,x:ord)
+
 -- | Call 'select' but return the result as a list.
-selectList :: (PersistEntity val, PersistBackend m, Monad m)
+selectList :: (PersistEntity val, PersistBackend b m)
            => [Filter val]
-           -> [Order val]
-           -> Int -- ^ limit
-           -> Int -- ^ offset
-           -> m [(Key val, val)]
-selectList a b c d = do
-    res <- run $ select a b c d ==<< consume
+           -> [SelectOpt val]
+           -> b m [(Key b val, val)]
+selectList a b = do
+    res <- run $ selectEnum a b ==<< consume
     case res of
-        Left e -> error $ show e
+        Left e -> Trans.liftIO . E.throwIO $ PersistError $ show e
         Right x -> return x
 
 data EntityDef = EntityDef
     { entityName    :: String
     , entityAttribs :: [String]
-    , entityColumns :: [(String, String, [String])] -- ^ name, type, attribs
-    , entityUniques :: [(String, [String])] -- ^ name, columns
+    , entityColumns :: [ColumnDef]
+    , entityUniques :: [UniqueDef]
     , entityDerives :: [String]
     }
     deriving Show
 
-instance Lift EntityDef where
-    lift (EntityDef a b c d e) = do
-        x <- [|EntityDef|]
-        a' <- lift a
-        b' <- lift b
-        c' <- lift c
-        d' <- lift d
-        e' <- lift e
-        return $ x `AppE` a' `AppE` b' `AppE` c' `AppE` d' `AppE` e'
+type ColumnName = String
+type ColumnType = String
+
+data ColumnDef = ColumnDef -- FIXME rename to FieldDef? Also, do we need a columnHaskellName and columnDBName?
+    { columnName    :: ColumnName
+    , columnType    :: ColumnType
+    , columnAttribs :: [String]
+    }
+    deriving Show
+
+data UniqueDef = UniqueDef
+    { uniqueName    :: String
+    , uniqueColumns :: [ColumnName]
+    }
+    deriving Show
 
 data PersistFilter = Eq | Ne | Gt | Lt | Ge | Le | In | NotIn
+                   | BackendSpecificFilter String
     deriving (Read, Show)
 
-instance Lift PersistFilter where
-    lift Eq = [|Eq|]
-    lift Ne = [|Ne|]
-    lift Gt = [|Gt|]
-    lift Lt = [|Lt|]
-    lift Ge = [|Ge|]
-    lift Le = [|Le|]
-    lift In = [|In|]
-    lift NotIn = [|NotIn|]
+class PersistEntity a => DeleteCascade a b where
+    deleteCascade :: PersistBackend b m => Key b a -> b m ()
 
-data PersistOrder = Asc | Desc
-    deriving (Read, Show)
-
-instance Lift PersistOrder where
-    lift Asc = [|Asc|]
-    lift Desc = [|Desc|]
-
-class PersistEntity a => DeleteCascade a where
-    deleteCascade :: PersistBackend m => Key a -> m ()
-
-deleteCascadeWhere :: (DeleteCascade a, PersistBackend m)
-                   => [Filter a] -> m ()
+deleteCascadeWhere :: (DeleteCascade a b, PersistBackend b m)
+                   => [Filter a] -> b m ()
 deleteCascadeWhere filts = do
     res <- run $ selectKeys filts $ Continue iter
     case res of
-        Left e -> error $ show e
+        Left e -> Trans.liftIO . E.throwIO $ PersistError $ show e
         Right () -> return ()
   where
     iter EOF = Iteratee $ return $ Yield () EOF
@@ -431,16 +566,24 @@ deleteCascadeWhere filts = do
         mapM_ deleteCascade keys
         return $ Continue iter
 
-data PersistException = PersistMarshalException String
-    deriving (Show, Typeable)
-instance E.Exception PersistException
 
-data PersistUpdate = Update | Add | Subtract | Multiply | Divide
-    deriving (Read, Show)
+data PersistUpdate = Assign | Add | Subtract | Multiply | Divide -- FIXME need something else here
+    deriving (Read, Show, Enum, Bounded)
 
-instance Lift PersistUpdate where
-    lift Update = [|Update|]
-    lift Add = [|Add|]
-    lift Subtract = [|Subtract|]
-    lift Multiply = [|Multiply|]
-    lift Divide = [|Divide|]
+instance PersistField PersistValue where
+    toPersistValue = id
+    fromPersistValue = Right
+    sqlType _ = SqlInteger -- since PersistValue should only be used like this for keys, which in SQL are Int64
+
+-- | Represents a value containing all the configuration options for a specific
+-- backend. This abstraction makes it easier to write code that can easily swap
+-- backends.
+class PersistConfig c where
+    type PersistConfigBackend c :: (* -> *) -> * -> *
+    type PersistConfigPool c
+    loadConfig :: TextObject -> Either String c
+    -- | I really don't want Applicative here, but it's necessary for Mongo.
+    withPool :: (Applicative m, MonadControlIO m) => c -> (PersistConfigPool c -> m a) -> m a
+    runPool :: MonadControlIO m => c -> PersistConfigBackend c m a
+            -> PersistConfigPool c
+            -> m a
