@@ -1,5 +1,9 @@
-{-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE PackageImports #-}
+{-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 -- | This is a helper module for creating SQL backends. Regular users do not
 -- need to use this module.
@@ -19,7 +23,9 @@ module Database.Persist.GenericSql
     , runMigrationSilent
     , runMigrationUnsafe
     , migrate
-    , mkMigrate
+    , commit
+    , rollback
+    , Key (..)
     ) where
 
 import Database.Persist.Base
@@ -27,7 +33,7 @@ import Data.List (intercalate)
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Reader
 import Control.Monad.Trans.Class (MonadTrans (..))
-import Database.Persist.Pool
+import Data.Pool
 import Control.Monad.Trans.Writer
 import System.IO
 import Database.Persist.GenericSql.Internal
@@ -35,33 +41,45 @@ import qualified Database.Persist.GenericSql.Raw as R
 import Database.Persist.GenericSql.Raw (SqlPersist (..))
 import Control.Monad (liftM, unless)
 import Data.Enumerator (Stream (..), Iteratee (..), Step (..))
-import Language.Haskell.TH.Syntax hiding (lift)
-import Control.Monad.Invert (MonadInvertIO, onException)
-import Control.Exception (toException)
+import Control.Monad.IO.Control (MonadControlIO)
+import Control.Exception.Control (onException)
+import Control.Exception (throw, toException)
+import Data.Text (Text, pack, unpack, snoc)
+import qualified Data.Text.IO
+import Web.PathPieces (SinglePiece (..))
+import qualified Data.Text.Read
 
 type ConnectionPool = Pool Connection
 
-withStmt' :: MonadInvertIO m => String -> [PersistValue]
+instance SinglePiece (Key SqlPersist entity) where
+    toSinglePiece (Key (PersistInt64 i)) = toSinglePiece i
+    toSinglePiece k = throw $ PersistInvalidField $ "Invalid Key: " ++ show k
+    fromSinglePiece t =
+        case Data.Text.Read.signed Data.Text.Read.decimal t of
+            Right (i, "") -> Just $ Key $ PersistInt64 i
+            _ -> Nothing
+
+withStmt' :: MonadControlIO m => Text -> [PersistValue]
          -> (RowPopper (SqlPersist m) -> SqlPersist m a) -> SqlPersist m a
 withStmt' = R.withStmt
 
-execute' :: MonadIO m => String -> [PersistValue] -> SqlPersist m ()
+execute' :: MonadIO m => Text -> [PersistValue] -> SqlPersist m ()
 execute' = R.execute
 
-runSqlPool :: MonadInvertIO m => SqlPersist m a -> Pool Connection -> m a
+runSqlPool :: MonadControlIO m => SqlPersist m a -> Pool Connection -> m a
 runSqlPool r pconn = withPool' pconn $ runSqlConn r
 
-runSqlConn :: MonadInvertIO m => SqlPersist m a -> Connection -> m a
+runSqlConn :: MonadControlIO m => SqlPersist m a -> Connection -> m a
 runSqlConn (SqlPersist r) conn = do
     let getter = R.getStmt' conn
     liftIO $ begin conn getter
     x <- onException
             (runReaderT r conn)
-            (liftIO $ rollback conn getter)
-    liftIO $ commit conn getter
+            (liftIO $ rollbackC conn getter)
+    liftIO $ commitC conn getter
     return x
 
-instance MonadInvertIO m => PersistBackend (SqlPersist m) where
+instance MonadControlIO m => PersistBackend SqlPersist m where
     insert val = do
         conn <- SqlPersist ask
         let esql = insertSql conn (rawTableName t) (map fst3 $ tableColumns t)
@@ -75,16 +93,15 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
                     withStmt' sql2 [] $ \pop -> do
                         Just [PersistInt64 i] <- pop
                         return i
-        return $ toPersistKey i
+        return $ Key $ PersistInt64 i
       where
-        fst3 (x, _, _) = x
         t = entityDef val
         vals = map toPersistValue $ toPersistFields val
 
     replace k val = do
         conn <- SqlPersist ask
         let t = entityDef val
-        let sql = concat
+        let sql = pack $ concat
                 [ "UPDATE "
                 , escapeName conn (rawTableName t)
                 , " SET "
@@ -92,55 +109,57 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
                 , " WHERE id=?"
                 ]
         execute' sql $ map toPersistValue (toPersistFields val)
-                       ++ [PersistInt64 $ fromPersistKey k]
+                       ++ [unKey k]
       where
         go conn x = escapeName conn x ++ "=?"
-        fst3 (x, _, _) = x
 
     get k = do
         conn <- SqlPersist ask
         let t = entityDef $ dummyFromKey k
         let cols = intercalate ","
                  $ map (\(x, _, _) -> escapeName conn x) $ tableColumns t
-        let sql = concat
+        let sql = pack $ concat
                 [ "SELECT "
                 , cols
                 , " FROM "
                 , escapeName conn $ rawTableName t
                 , " WHERE id=?"
                 ]
-        withStmt' sql [PersistInt64 $ fromPersistKey k] $ \pop -> do
+        withStmt' sql [unKey k] $ \pop -> do
             res <- pop
             case res of
                 Nothing -> return Nothing
                 Just vals ->
                     case fromPersistValues vals of
-                        Left e -> error $ "get " ++ showPersistKey k ++ ": " ++ e
+                        Left e -> error $ "get " ++ show (unKey k) ++ ": " ++ e
                         Right v -> return $ Just v
 
     count filts = do
         conn <- SqlPersist ask
         let wher = if null filts
                     then ""
-                    else " WHERE " ++
-                         intercalate " AND " (map (filterClause conn) filts)
-        let sql = concat
+                    else filterClause False conn filts
+        let sql = pack $ concat
                 [ "SELECT COUNT(*) FROM "
                 , escapeName conn $ rawTableName t
                 , wher
                 ]
-        withStmt' sql (getFiltsValues filts) $ \pop -> do
+        withStmt' sql (getFiltsValues conn filts) $ \pop -> do
             Just [PersistInt64 i] <- pop
             return $ fromIntegral i
       where
         t = entityDef $ dummyFromFilts filts
 
-    select filts ords limit offset =
+    selectEnum filts opts =
         Iteratee . start
       where
+        limit  = fst3 $ limitOffsetOrder opts
+        offset = snd3 $ limitOffsetOrder opts
+        orders = third3 $ limitOffsetOrder opts
+
         start x = do
             conn <- SqlPersist ask
-            withStmt' (sql conn) (getFiltsValues filts) $ loop x
+            withStmt' (sql conn) (getFiltsValues conn filts) $ loop x
         loop (Continue k) pop = do
             res <- pop
             case res of
@@ -148,30 +167,24 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
                 Just vals -> do
                     case fromPersistValues' vals of
                         Left s -> return $ Error $ toException
-                                $ PersistMarshalException s
+                                $ PersistMarshalError s
                         Right row -> do
                             step <- runIteratee $ k $ Chunks [row]
                             loop step pop
         loop step _ = return step
         t = entityDef $ dummyFromFilts filts
-        orderClause conn o =
-            escapeName conn (getFieldName t $ persistOrderToFieldName o)
-                        ++ case persistOrderToOrder o of
-                                            Asc -> ""
-                                            Desc -> " DESC"
         fromPersistValues' (PersistInt64 x:xs) = do
             case fromPersistValues xs of
                 Left e -> Left e
-                Right xs' -> Right (toPersistKey x, xs')
+                Right xs' -> Right (Key $ PersistInt64 x, xs')
         fromPersistValues' _ = Left "error in fromPersistValues'"
         wher conn = if null filts
                     then ""
-                    else " WHERE " ++
-                         intercalate " AND " (map (filterClause conn) filts)
-        ord conn = if null ords
-                    then ""
-                    else " ORDER BY " ++
-                         intercalate "," (map (orderClause conn) ords)
+                    else filterClause False conn filts
+        ord conn =
+            case map (orderClause False conn) orders of
+                [] -> ""
+                ords -> " ORDER BY " ++ intercalate "," ords
         lim conn = case (limit, offset) of
                 (0, 0) -> ""
                 (0, _) -> ' ' : noLimit conn
@@ -179,9 +192,9 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
         off = if offset == 0
                     then ""
                     else " OFFSET " ++ show offset
-        cols conn = intercalate "," $ "id"
+        cols conn = intercalate "," $ (unRawName $ rawTableIdName t)
                    : (map (\(x, _, _) -> escapeName conn x) $ tableColumns t)
-        sql conn = concat
+        sql conn = pack $ concat
             [ "SELECT "
             , cols conn
             , " FROM "
@@ -192,29 +205,27 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
             , off
             ]
 
-
     selectKeys filts =
         Iteratee . start
       where
         start x = do
             conn <- SqlPersist ask
-            withStmt' (sql conn) (getFiltsValues filts) $ loop x
+            withStmt' (sql conn) (getFiltsValues conn filts) $ loop x
         loop (Continue k) pop = do
             res <- pop
             case res of
                 Nothing -> return $ Continue k
                 Just [PersistInt64 i] -> do
-                    step <- runIteratee $ k $ Chunks [toPersistKey i]
+                    step <- runIteratee $ k $ Chunks [Key $ PersistInt64 i]
                     loop step pop
-                Just y -> return $ Error $ toException $ PersistMarshalException
+                Just y -> return $ Error $ toException $ PersistMarshalError
                         $ "Unexpected in selectKeys: " ++ show y
         loop step _ = return step
         t = entityDef $ dummyFromFilts filts
         wher conn = if null filts
                     then ""
-                    else " WHERE " ++
-                         intercalate " AND " (map (filterClause conn) filts)
-        sql conn = concat
+                    else filterClause False conn filts
+        sql conn = pack $ concat
             [ "SELECT id FROM "
             , escapeName conn $ rawTableName t
             , wher conn
@@ -222,10 +233,10 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
 
     delete k = do
         conn <- SqlPersist ask
-        execute' (sql conn) [PersistInt64 $ fromPersistKey k]
+        execute' (sql conn) [unKey k]
       where
         t = entityDef $ dummyFromKey k
-        sql conn = concat
+        sql conn = pack $ concat
             [ "DELETE FROM "
             , escapeName conn $ rawTableName t
             , " WHERE id=?"
@@ -236,14 +247,13 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
         let t = entityDef $ dummyFromFilts filts
         let wher = if null filts
                     then ""
-                    else " WHERE " ++
-                         intercalate " AND " (map (filterClause conn) filts)
-            sql = concat
+                    else filterClause False conn filts
+            sql = pack $ concat
                 [ "DELETE FROM "
                 , escapeName conn $ rawTableName t
                 , wher
                 ]
-        execute' sql $ getFiltsValues filts
+        execute' sql $ getFiltsValues conn filts
 
     deleteBy uniq = do
         conn <- SqlPersist ask
@@ -252,7 +262,7 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
         t = entityDef $ dummyFromUnique uniq
         go = map (getFieldName t) . persistUniqueToFieldNames
         go' conn x = escapeName conn x ++ "=?"
-        sql conn = concat
+        sql conn = pack $ concat
             [ "DELETE FROM "
             , escapeName conn $ rawTableName t
             , " WHERE "
@@ -262,13 +272,13 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
     update _ [] = return ()
     update k upds = do
         conn <- SqlPersist ask
-        let go'' n Update = n ++ "=?"
+        let go'' n Assign = n ++ "=?"
             go'' n Add = n ++ '=' : n ++ "+?"
             go'' n Subtract = n ++ '=' : n ++ "-?"
             go'' n Multiply = n ++ '=' : n ++ "*?"
             go'' n Divide = n ++ '=' : n ++ "/?"
         let go' (x, pu) = go'' (escapeName conn x) pu
-        let sql = concat
+        let sql = pack $ concat
                 [ "UPDATE "
                 , escapeName conn $ rawTableName t
                 , " SET "
@@ -276,11 +286,11 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
                 , " WHERE id=?"
                 ]
         execute' sql $
-            map persistUpdateToValue upds ++ [PersistInt64 $ fromPersistKey k]
+            map updatePersistValue upds ++ [unKey k]
       where
         t = entityDef $ dummyFromKey k
-        go x = ( getFieldName t $ persistUpdateToFieldName x
-               , persistUpdateToUpdate x
+        go x = ( getFieldName t $ updateFieldName x
+               , updateUpdate x
                )
 
     updateWhere _ [] = return ()
@@ -288,34 +298,33 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
         conn <- SqlPersist ask
         let wher = if null filts
                     then ""
-                    else " WHERE " ++
-                         intercalate " AND " (map (filterClause conn) filts)
-        let sql = concat
+                    else filterClause False conn filts
+        let sql = pack $ concat
                 [ "UPDATE "
                 , escapeName conn $ rawTableName t
                 , " SET "
                 , intercalate "," $ map (go' conn . go) upds
                 , wher
                 ]
-        let dat = map persistUpdateToValue upds ++ getFiltsValues filts
+        let dat = map updatePersistValue upds ++ getFiltsValues conn filts
         execute' sql dat
       where
         t = entityDef $ dummyFromFilts filts
-        go'' n Update = n ++ "=?"
+        go'' n Assign = n ++ "=?"
         go'' n Add = n ++ '=' : n ++ "+?"
         go'' n Subtract = n ++ '=' : n ++ "-?"
         go'' n Multiply = n ++ '=' : n ++ "*?"
         go'' n Divide = n ++ '=' : n ++ "/?"
         go' conn (x, pu) = go'' (escapeName conn x) pu
-        go x = ( getFieldName t $ persistUpdateToFieldName x
-               , persistUpdateToUpdate x
+        go x = ( getFieldName t $ updateFieldName x
+               , updateUpdate x
                )
 
     getBy uniq = do
         conn <- SqlPersist ask
-        let cols = intercalate "," $ "id"
+        let cols = intercalate "," $ (unRawName $ rawTableIdName t)
                  : (map (\(x, _, _) -> escapeName conn x) $ tableColumns t)
-        let sql = concat
+        let sql = pack $ concat
                 [ "SELECT "
                 , cols
                 , " FROM "
@@ -330,7 +339,7 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
                 Just (PersistInt64 k:vals) ->
                     case fromPersistValues vals of
                         Left s -> error s
-                        Right x -> return $ Just (toPersistKey k, x)
+                        Right x -> return $ Just (Key $ PersistInt64 k, x)
                 Just _ -> error "Database.Persist.GenericSql: Bad list in getBy"
       where
         sqlClause conn =
@@ -339,96 +348,14 @@ instance MonadInvertIO m => PersistBackend (SqlPersist m) where
         t = entityDef $ dummyFromUnique uniq
         toFieldNames' = map (getFieldName t) . persistUniqueToFieldNames
 
-dummyFromUnique :: Unique v -> v
+dummyFromUnique :: Unique v b -> v
 dummyFromUnique _ = error "dummyFromUnique"
 
-getFieldName :: EntityDef -> String -> RawName
-getFieldName t s = rawFieldName $ tableColumn t s
-
-tableColumn :: EntityDef -> String -> (String, String, [String])
-tableColumn t s = go $ entityColumns t
-  where
-    go [] = error $ "Unknown table column: " ++ s
-    go ((x, y, z):rest)
-        | x == s = (x, y, z)
-        | otherwise = go rest
-
-dummyFromKey :: Key v -> v
+dummyFromKey :: Key SqlPersist v -> v
 dummyFromKey _ = error "dummyFromKey"
 
-filterClause :: PersistEntity val => Connection -> Filter val -> String
-filterClause conn f =
-    case (isNull, persistFilterToFilter f, varCount) of
-        (True, Eq, _) -> name ++ " IS NULL"
-        (True, Ne, _) -> name ++ " IS NOT NULL"
-        (False, Ne, _) -> concat
-            [ "("
-            , name
-            , " IS NULL OR "
-            , name
-            , "<>?)"
-            ]
-        -- We use 1=2 (and below 1=1) to avoid using TRUE and FALSE, since
-        -- not all databases support those words directly.
-        (_, In, 0) -> "1=2"
-        (False, In, _) -> name ++ " IN " ++ qmarks
-        (True, In, _) -> concat
-            [ "("
-            , name
-            , " IS NULL OR "
-            , name
-            , " IN "
-            , qmarks
-            , ")"
-            ]
-        (_, NotIn, 0) -> "1=1"
-        (False, NotIn, _) -> concat
-            [ "("
-            , name
-            , " IS NULL OR "
-            , name
-            , " NOT IN "
-            , qmarks
-            , ")"
-            ]
-        (True, NotIn, _) -> concat
-            [ "("
-            , name
-            , " IS NOT NULL AND "
-            , name
-            , " NOT IN "
-            , qmarks
-            , ")"
-            ]
-        _ -> name ++ showSqlFilter (persistFilterToFilter f) ++ "?"
-  where
-    isNull = any (== PersistNull)
-           $ either return id
-           $ persistFilterToValue f
-    t = entityDef $ dummyFromFilts [f]
-    name = escapeName conn $ getFieldName t $ persistFilterToFieldName f
-    qmarks = case persistFilterToValue f of
-                Left _ -> "?"
-                Right x ->
-                    let x' = filter (/= PersistNull) x
-                     in '(' : intercalate "," (map (const "?") x') ++ ")"
-    varCount = case persistFilterToValue f of
-                Left _ -> 1
-                Right x -> length x
-    showSqlFilter Eq = "="
-    showSqlFilter Ne = "<>"
-    showSqlFilter Gt = ">"
-    showSqlFilter Lt = "<"
-    showSqlFilter Ge = ">="
-    showSqlFilter Le = "<="
-    showSqlFilter In = " IN "
-    showSqlFilter NotIn = " NOT IN "
 
-dummyFromFilts :: [Filter v] -> v
-dummyFromFilts _ = error "dummyFromFilts"
-
-
-type Sql = String
+type Sql = Text
 
 -- Bool indicates if the Sql is safe
 type CautiousMigration = [(Bool, Sql)]
@@ -439,9 +366,9 @@ unsafeSql = allSql . filter fst
 safeSql :: CautiousMigration -> [Sql]
 safeSql = allSql . filter (not . fst)
 
-type Migration m = WriterT [String] (WriterT CautiousMigration m) ()
+type Migration m = WriterT [Text] (WriterT CautiousMigration m) ()
 
-parseMigration :: Monad m => Migration m -> m (Either [String] CautiousMigration)
+parseMigration :: Monad m => Migration m -> m (Either [Text] CautiousMigration)
 parseMigration =
     liftM go . runWriterT . execWriterT
   where
@@ -453,35 +380,35 @@ parseMigration' :: Monad m => Migration m -> m (CautiousMigration)
 parseMigration' m = do
   x <- parseMigration m
   case x of
-      Left errs -> error $ unlines errs
+      Left errs -> error $ unlines $ map unpack errs
       Right sql -> return sql
 
-printMigration :: MonadInvertIO m => Migration (SqlPersist m) -> SqlPersist m ()
+printMigration :: MonadControlIO m => Migration (SqlPersist m) -> SqlPersist m ()
 printMigration m = do
   mig <- parseMigration' m
-  mapM_ (liftIO . putStrLn) (allSql mig)
+  mapM_ (liftIO . Data.Text.IO.putStrLn . flip snoc ';') (allSql mig)
 
-getMigration :: MonadInvertIO m => Migration (SqlPersist m) -> SqlPersist m [Sql]
+getMigration :: MonadControlIO m => Migration (SqlPersist m) -> SqlPersist m [Sql]
 getMigration m = do
   mig <- parseMigration' m
   return $ allSql mig
 
-runMigration :: MonadInvertIO m
+runMigration :: MonadControlIO m
              => Migration (SqlPersist m)
              -> SqlPersist m ()
 runMigration m = runMigration' m False >> return ()
 
 -- | Same as 'runMigration', but returns a list of the SQL commands executed
 -- instead of printing them to stderr.
-runMigrationSilent :: MonadInvertIO m
+runMigrationSilent :: MonadControlIO m
                    => Migration (SqlPersist m)
-                   -> SqlPersist m [String]
+                   -> SqlPersist m [Text]
 runMigrationSilent m = runMigration' m True
 
-runMigration' :: MonadInvertIO m
+runMigration' :: MonadControlIO m
               => Migration (SqlPersist m)
               -> Bool -- ^ is silent?
-              -> SqlPersist m [String]
+              -> SqlPersist m [Text]
 runMigration' m silent = do
     mig <- parseMigration' m
     case unsafeSql mig of
@@ -489,23 +416,23 @@ runMigration' m silent = do
         errs -> error $ concat
             [ "\n\nDatabase migration: manual intervention required.\n"
             , "The following actions are considered unsafe:\n\n"
-            , unlines $ map (\s -> "    " ++ s ++ ";") $ errs
+            , unlines $ map (\s -> "    " ++ unpack s ++ ";") $ errs
             ]
 
-runMigrationUnsafe :: MonadInvertIO m
+runMigrationUnsafe :: MonadControlIO m
                    => Migration (SqlPersist m)
                    -> SqlPersist m ()
 runMigrationUnsafe m = do
     mig <- parseMigration' m
     mapM_ (executeMigrate False) $ allSql mig
 
-executeMigrate :: MonadIO m => Bool -> String -> SqlPersist m String
+executeMigrate :: MonadIO m => Bool -> Text -> SqlPersist m Text
 executeMigrate silent s = do
-    unless silent $ liftIO $ hPutStrLn stderr $ "Migrating: " ++ s
+    unless silent $ liftIO $ hPutStrLn stderr $ "Migrating: " ++ unpack s
     execute' s []
     return s
 
-migrate :: (MonadInvertIO m, PersistEntity val)
+migrate :: (MonadControlIO m, PersistEntity val)
         => val
         -> Migration (SqlPersist m)
 migrate val = do
@@ -514,37 +441,19 @@ migrate val = do
     res <- liftIO $ migrateSql conn getter val
     either tell (lift . tell) res
 
-getFiltsValues :: PersistEntity val => [Filter val] -> [PersistValue]
-getFiltsValues =
-    concatMap $ go . persistFilterToValue
-  where
-    go (Left PersistNull) = []
-    go (Left x) = [x]
-    go (Right xs) = filter (/= PersistNull) xs
+updatePersistValue :: Update v -> PersistValue
+updatePersistValue (Update _ v _) = toPersistValue v
 
--- | Creates a single function to perform all migrations for the entities
--- defined here. One thing to be aware of is dependencies: if you have entities
--- with foreign references, make sure to place those definitions after the
--- entities they reference.
-mkMigrate :: String -> [EntityDef] -> Q [Dec]
-mkMigrate fun defs = do
-    body' <- body
-    return
-        [ SigD (mkName fun) typ
-        , FunD (mkName fun) [Clause [] (NormalB body') []]
-        ]
-  where
-    typ = ForallT [PlainTV $ mkName "m"]
-            [ ClassP ''MonadInvertIO [VarT $ mkName "m"]
-            ]
-            $ ConT ''Migration `AppT` (ConT ''SqlPersist `AppT` VarT (mkName "m"))
-    body =
-        case defs of
-            [] -> [|return ()|]
-            _ -> DoE `fmap` mapM toStmt defs
-    toStmt ed = do
-        let n = entityName ed
-        u <- [|undefined|]
-        m <- [|migrate|]
-        let u' = SigE u $ ConT $ mkName n
-        return $ NoBindS $ m `AppE` u'
+-- | Perform a database commit.
+commit :: MonadIO m => SqlPersist m ()
+commit = do
+    conn <- SqlPersist ask
+    let getter = R.getStmt' conn
+    liftIO $ commitC conn getter >> begin conn getter
+
+-- | Perform a database rollback.
+rollback :: MonadIO m => SqlPersist m ()
+rollback = do
+    conn <- SqlPersist ask
+    let getter = R.getStmt' conn
+    liftIO $ rollbackC conn getter >> begin conn getter
